==> kafkakit-0.4.2.jar: kafkakit/avro.clj <==
(ns kafkakit.avro
  "Convert Clojure data structures to Avro-compatible Java classes (not Avro
  binary format!) back and forth in order to be able to use Schema Registry's
  serializers with Kafka and Clojure.

  Source: https://github.com/konukhov/kfk-avro-bridge

  References:

    * https://github.com/confluentinc/schema-registry/blob/master/avro-serializer/src/main/java/io/confluent/kafka/serializers/KafkaAvroSerializer.java
  "
  (:refer-clojure :exclude [bytes?])
  (:require
    [clj-time.coerce :as c])
  (:import
    [java.nio ByteBuffer]
    [org.apache.avro.generic GenericData]
    [org.apache.avro.generic GenericData$Array]
    [org.apache.avro.generic GenericData$EnumSymbol]
    [org.apache.avro.generic GenericData$Fixed]
    [org.apache.avro.generic GenericData$Record]
    [org.apache.avro.util Utf8]
    [org.apache.avro Schema Schema$Type]))

  (defn- bytes?
    "Compatibility with Clojure < 1.9"
    [x]
    (if (nil? x)
      false
      (-> x class .getComponentType (= Byte/TYPE))))

  (defn- throw-invalid-type
    [^Schema schema obj]
    (throw (Exception. (format "Value `%s` cannot be cast to `%s` schema"
                               (str obj)
                               (.toString schema)))))

  (defn parse-schema
    "A little helper for parsing schemas"
    [json]
    (Schema/parse json))

  (defn logical-name [schema]
    (when (instance? Schema schema)
      (some->> schema
               .getLogicalType
               .getName)))

  (def early-epoch-millis
    11882926000) ;; Just a random time earlier than we will deal with
                 ;; Mon 18 May 1970 07:48:46

  (defn epoch-days-to-epoch-millis
    [epoch-days]
    (* epoch-days 86400000))

  (defn fix-epoch-millis
    "Due to brain dead Kafka Connect date handling, or lack thereof,
     sometimes input here must be epoch days and must be converted."
    [maybe-epoch-millis]
    (let [epoch-millis (if (< maybe-epoch-millis early-epoch-millis)
                         (epoch-days-to-epoch-millis maybe-epoch-millis)
                         maybe-epoch-millis)]
      (-> epoch-millis c/to-date-time c/to-long)))

  (defn ->java
    "Converts a Clojure data structure to an Avro-compatible
     Java object. Avro `Schema` must be provided."
    [schema obj]
    (condp = (and (instance? Schema schema) (.getType schema))
      Schema$Type/NULL
      (if (nil? obj)
        nil
        (throw-invalid-type schema obj))

      Schema$Type/INT
      (if (and (integer? obj) (<= Integer/MIN_VALUE obj Integer/MAX_VALUE))
        (int obj)
        (throw-invalid-type schema obj))

      Schema$Type/LONG
      (cond
        (= "timestamp-millis" (logical-name schema))
        (fix-epoch-millis obj)
        (and (integer? obj) (<= Long/MIN_VALUE obj Long/MAX_VALUE)) (long obj)
        :default (throw-invalid-type schema obj))

      Schema$Type/FLOAT
      (if (float? obj)
        (float obj)
        (throw-invalid-type schema obj))

      Schema$Type/DOUBLE
      (if (float? obj)
        (double obj)
        (throw-invalid-type schema obj))

      Schema$Type/BOOLEAN
      (if (instance? Boolean obj) ;; boolean? added only in 1.9 :(
        obj
        (throw-invalid-type schema obj))

      Schema$Type/STRING
      (cond
        (string? obj) obj
        (instance? java.util.UUID obj) (str obj)
        :default (throw-invalid-type schema obj))

      Schema$Type/BYTES
      (if (bytes? obj)
        (doto (ByteBuffer/allocate (count obj))
          (.put obj)
          (.position 0))
        (throw-invalid-type schema obj))

      Schema$Type/ARRAY ;; TODO Exception for complex type
      (let [f (partial ->java (.getElementType schema))]
        (GenericData$Array. schema (map f obj)))

      Schema$Type/FIXED
      (if (and (bytes? obj) (= (count obj) (.getFixedSize schema)))
        (GenericData$Fixed. schema obj)
        (throw-invalid-type schema obj))

      Schema$Type/ENUM
      (let [enum obj
            enums (into #{} (.getEnumSymbols schema))]
        (if (contains? enums enum)
          (GenericData$EnumSymbol. schema enum)
          (throw-invalid-type schema enum)))

      Schema$Type/MAP ;; TODO Exception for complex type
      (zipmap (map name (keys obj))
            (map (partial ->java (.getValueType schema))
                 (vals obj)))

    Schema$Type/UNION
    (let [[val matched]
          (reduce (fn [_ schema]
                    (try
                      (reduced [(->java schema obj) true])
                      (catch Exception _
                        [nil false])))
                  [nil false]
                  (.getTypes schema))]
      (if matched
        val
        (throw-invalid-type schema obj)))

    Schema$Type/RECORD
    (reduce-kv
     (fn [record k v]
       (let [k (name k)
             s (some-> (.getField schema k)
                       (as-> f (.schema f)))]
         (doto record
           (.put k (->java (or s k) v)))))
     (GenericData$Record. schema)
     obj)

    (throw (Exception. (format "Field `%s` is not in schema" schema)))))

(defn ->clj
  "Parses deserialized Avro object into a Clojure data structure."
  [msg]
  (condp instance? msg
    Utf8
    (str msg)

    java.nio.ByteBuffer
    (.array msg)

    GenericData$Array
    (into [] (map ->clj msg))

    GenericData$Fixed
    (.bytes msg)

    GenericData$EnumSymbol
    (str msg)

    java.util.Map
    (zipmap (map (comp str) (keys msg))
            (map ->clj (vals msg)))

    GenericData$Record
    ;; Transients are making this slower, I wonder why?
    (loop [fields (seq (.. msg getSchema getFields)) record {}]
      (if-let [f (first fields)]
        (let [n (.name f)
              v (->clj (.get msg n))
              k (-> n keyword)]
          (recur (rest fields)
                 (assoc record k v)))
        record))

    msg))

==> kafkakit-0.4.3.jar: kafkakit/avro.clj <==
(ns kafkakit.avro
  "Convert Clojure data structures to Avro-compatible Java classes (not Avro
  binary format!) back and forth in order to be able to use Schema Registry's
  serializers with Kafka and Clojure.

  Source: https://github.com/konukhov/kfk-avro-bridge

  References:

    * https://github.com/confluentinc/schema-registry/blob/master/avro-serializer/src/main/java/io/confluent/kafka/serializers/KafkaAvroSerializer.java
  "
  (:refer-clojure :exclude [bytes?])
  (:require
    [clj-time.coerce :as c])
  (:import
    [java.nio ByteBuffer]
    [org.apache.avro.generic GenericData]
    [org.apache.avro.generic GenericData$Array]
    [org.apache.avro.generic GenericData$EnumSymbol]
    [org.apache.avro.generic GenericData$Fixed]
    [org.apache.avro.generic GenericData$Record]
    [org.apache.avro.util Utf8]
    [org.apache.avro Schema Schema$Type]))

  (defn- bytes?
    "Compatibility with Clojure < 1.9"
    [x]
    (if (nil? x)
      false
      (-> x class .getComponentType (= Byte/TYPE))))

  (defn- throw-invalid-type
    [^Schema schema obj]
    (throw (Exception. (format "Value `%s` cannot be cast to `%s` schema"
                               (str obj)
                               (.toString schema)))))

  (defn parse-schema
    "A little helper for parsing schemas"
    [json]
    (Schema/parse json))

  (defn logical-name [schema]
    (when (instance? Schema schema)
      (some->> schema
               .getLogicalType
               .getName)))

  (def early-epoch-millis
    11882926000) ;; Just a random time earlier than we will deal with
                 ;; Mon 18 May 1970 07:48:46

  (defn epoch-days-to-epoch-millis
    [epoch-days]
    (* epoch-days 86400000))

  (defn fix-epoch-millis
    "Due to brain dead Kafka Connect date handling, or lack thereof,
     sometimes input here must be epoch days and must be converted."
    [maybe-epoch-millis]
    (let [epoch-millis (if (and (integer? maybe-epoch-millis) ;; may be integer or joda.time.DateTime
                                (< maybe-epoch-millis early-epoch-millis))
                         (epoch-days-to-epoch-millis maybe-epoch-millis)
                         maybe-epoch-millis)]
      (-> epoch-millis c/to-date-time c/to-long)))

  (defn ->java
    "Converts a Clojure data structure to an Avro-compatible
     Java object. Avro `Schema` must be provided."
    [schema obj]
    (condp = (and (instance? Schema schema) (.getType schema))
      Schema$Type/NULL
      (if (nil? obj)
        nil
        (throw-invalid-type schema obj))

      Schema$Type/INT
      (if (and (integer? obj) (<= Integer/MIN_VALUE obj Integer/MAX_VALUE))
        (int obj)
        (throw-invalid-type schema obj))

      Schema$Type/LONG
      (cond
        (= "timestamp-millis" (logical-name schema))
        (fix-epoch-millis obj)
        (and (integer? obj) (<= Long/MIN_VALUE obj Long/MAX_VALUE)) (long obj)
        :default (throw-invalid-type schema obj))

      Schema$Type/FLOAT
      (if (float? obj)
        (float obj)
        (throw-invalid-type schema obj))

      Schema$Type/DOUBLE
      (if (float? obj)
        (double obj)
        (throw-invalid-type schema obj))

      Schema$Type/BOOLEAN
      (if (instance? Boolean obj) ;; boolean? added only in 1.9 :(
        obj
        (throw-invalid-type schema obj))

      Schema$Type/STRING
      (cond
        (string? obj) obj
        (instance? java.util.UUID obj) (str obj)
        :default (throw-invalid-type schema obj))

      Schema$Type/BYTES
      (if (bytes? obj)
        (doto (ByteBuffer/allocate (count obj))
          (.put obj)
          (.position 0))
        (throw-invalid-type schema obj))

      Schema$Type/ARRAY ;; TODO Exception for complex type
      (let [f (partial ->java (.getElementType schema))]
        (GenericData$Array. schema (map f obj)))

      Schema$Type/FIXED
      (if (and (bytes? obj) (= (count obj) (.getFixedSize schema)))
        (GenericData$Fixed. schema obj)
        (throw-invalid-type schema obj))

      Schema$Type/ENUM
      (let [enum obj
            enums (into #{} (.getEnumSymbols schema))]
        (if (contains? enums enum)
          (GenericData$EnumSymbol. schema enum)
          (throw-invalid-type schema enum)))

      Schema$Type/MAP ;; TODO Exception for complex type
      (zipmap (map name (keys obj))
            (map (partial ->java (.getValueType schema))
                 (vals obj)))

    Schema$Type/UNION
    (let [[val matched]
          (reduce (fn [_ schema]
                    (try
                      (reduced [(->java schema obj) true])
                      (catch Exception _
                        [nil false])))
                  [nil false]
                  (.getTypes schema))]
      (if matched
        val
        (throw-invalid-type schema obj)))

    Schema$Type/RECORD
    (reduce-kv
     (fn [record k v]
       (let [k (name k)
             s (some-> (.getField schema k)
                       (as-> f (.schema f)))]
         (doto record
           (.put k (->java (or s k) v)))))
     (GenericData$Record. schema)
     obj)

    (throw (Exception. (format "Field `%s` is not in schema" schema)))))

(defn ->clj
  "Parses deserialized Avro object into a Clojure data structure."
  [msg]
  (condp instance? msg
    Utf8
    (str msg)

    java.nio.ByteBuffer
    (.array msg)

    GenericData$Array
    (into [] (map ->clj msg))

    GenericData$Fixed
    (.bytes msg)

    GenericData$EnumSymbol
    (str msg)

    java.util.Map
    (zipmap (map (comp str) (keys msg))
            (map ->clj (vals msg)))

    GenericData$Record
    ;; Transients are making this slower, I wonder why?
    (loop [fields (seq (.. msg getSchema getFields)) record {}]
      (if-let [f (first fields)]
        (let [n (.name f)
              v (->clj (.get msg n))
              k (-> n keyword)]
          (recur (rest fields)
                 (assoc record k v)))
        record))

    msg))
